#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <descrip.h>

#include "db.h"

#include "vms_bdb.h"

static int size_of_int = sizeof(int);

static char *in_cstr(struct dsc$descriptor *s)
{
    int slen, ix;
    char *sptr, *res;
    if(s->dsc$b_class == DSC$K_CLASS_S || s->dsc$b_class == DSC$K_CLASS_D)
    {
        slen = s->dsc$w_length;
        sptr = s->dsc$a_pointer;
    }
    else if(s->dsc$b_class == DSC$K_CLASS_VS)
    {
        slen = *((short int *)s->dsc$a_pointer);
        sptr = s->dsc$a_pointer + sizeof(short int);
    }
    else
    {
        fprintf(stderr, "Unsupported descriptor class: %d\n", s->dsc$b_class);
        exit(44);
    }
    res = malloc(slen + 1);
    strncpy(res, sptr, slen);
    res[slen] = 0;
    /* trim traling whitespace */
    ix = slen - 1;
    while(ix >= 0 && (res[ix] == ' ' || res[ix] == 0))
    {
        res[ix] = 0;
        ix--;
    }
    return res;
}

int db_open(struct dsc$descriptor *dbnam)
{
    DB *db;
    char *dbnam2;
    int stat;
    dbnam2 = in_cstr(dbnam);
    stat = db_create(&db, NULL, 0);
    stat = db->open(db, NULL, dbnam2, NULL, DB_BTREE, DB_CREATE, 0664); /* DB_HASH is alternative to DB_BTREE */
    free(dbnam2);
    return (int)db;
}

int db_put(int *db, char *key, int *keylen, char *outbuf, int *outbuflen)
{
    DB *db2;
    DBT dbkey,dbvalue;
    int stat;
    db2 = (DB *)(*db);
    memset(&dbkey, 0, sizeof(DBT));
    dbkey.data = key;
    dbkey.size = (unsigned int)(*keylen);
    memset(&dbvalue, 0, sizeof(DBT));
    dbvalue.data = outbuf;
    dbvalue.size = (unsigned int)(*outbuflen);
    stat = db2->put(db2, NULL, &dbkey, &dbvalue, 0);
    return stat;
}

int db_get(int *db, char *key, int *keylen, char *inbuf, int *maxinbuflen, int *inbuflen)
{
    DB *db2;
    DBT dbkey, dbvalue;
    int stat;
    db2 = (DB *)(*db);
    memset(&dbkey, 0, sizeof(DBT));
    memset(&dbvalue, 0, sizeof(DBT));
    dbkey.data = key;
    dbkey.size = (unsigned int)(*keylen);
    dbvalue.size = (unsigned int)(*maxinbuflen);
    stat = db2->get(db2, NULL, &dbkey, &dbvalue, 0);
    if(stat == 0)
    {
        *inbuflen = dbvalue.size;
        memcpy(inbuf, dbvalue.data, dbvalue.size);
    }
    return stat;
}

int db_delete(int *db, char *key, int *keylen)
{
    DB *db2;
    DBT dbkey;
    int stat;
    db2 = (DB *)(*db);
    memset(&dbkey, 0, sizeof(DBT));
    dbkey.data = key;
    dbkey.size = (unsigned int)(*keylen);
    stat = db2->del(db2, NULL, &dbkey, 0);
    return stat;
}

int db_sk_put(int *db, struct dsc$descriptor *key, char *outbuf, int *outbuflen)
{
    int stat, key2len;
    char *key2;
    key2 = in_cstr(key);
    key2len = strlen(key2);
    stat = db_put(db, key2, &key2len, outbuf, outbuflen);
    free(key2);
    return stat;
}

int db_sk_get(int *db, struct dsc$descriptor *key, char *inbuf, int *maxinbuflen, int *inbuflen)
{
    int stat, key2len;
    char *key2;
    key2 = in_cstr(key);
    key2len = strlen(key2);
    stat = db_get(db, key2, &key2len, inbuf, maxinbuflen, inbuflen);
    free(key2);
    return stat;
}

int db_sk_delete(int *db, struct dsc$descriptor *key)
{
    int stat, key2len;
    char *key2;
    key2 = in_cstr(key);
    key2len = strlen(key2);
    stat = db_delete(db, key2, &key2len);
    free(key2);
    return stat;
}

int db_ik_put(int *db, int *key, char *outbuf, int *outbuflen)
{
    return db_put(db, (char *)key, &size_of_int, outbuf, outbuflen);
}

int db_ik_get(int *db, int *key, char *inbuf, int *maxinbuflen, int *inbuflen)
{
   return db_get(db, (char *)key, &size_of_int, inbuf, maxinbuflen, inbuflen);
}

int db_ik_delete(int *db, int *key)
{
   return db_delete(db, (char *)key, &size_of_int);
}

int db_cursor_open(int *db)
{
    DB *db2;
    DBC *it;
    int stat;
    char *key2;
    db2 = (DB *)(*db);
    stat = db2->cursor(db2, NULL, &it, 0);
    return (int)it;
}

int db_cursor_find(int *it, char *key, int *keylen, char *inbuf, int *maxinbuflen, int *inbuflen)
{
    DBC *it2;
    DBT dbkey, dbvalue;
    int stat;
    char *key2;
    it2 = (DBC *)(*it);
    memset(&dbkey, 0, sizeof(DBT));
    memset(&dbvalue, 0, sizeof(DBT));
    dbkey.data = key;
    dbkey.size = (unsigned int)(*keylen);
    dbvalue.size = (unsigned int)(*maxinbuflen);
    stat = it2->get(it2, &dbkey, &dbvalue, DB_SET);
    if(stat == 0)
    {
        *inbuflen = dbvalue.size;
        memcpy(inbuf, dbvalue.data, dbvalue.size);
    }
    return stat;
}

int db_cursor_next(int *it, char *key, int *maxkeylen, short int *keylen, char *inbuf, int *maxinbuflen, int *inbuflen)
{
    DBC *it2;
    DBT dbkey, dbvalue;
    int stat;
    it2 = (DBC *)(*it);
    memset(&dbkey, 0, sizeof(DBT));
    memset(&dbvalue, 0, sizeof(DBT));
    dbkey.size = (unsigned int)(*maxkeylen);
    dbvalue.size = (unsigned int)(*maxinbuflen);
    stat = it2->get(it2, &dbkey, &dbvalue, DB_NEXT);
    if(stat == 0)
    {
        memcpy(key, dbkey.data, dbkey.size);
        *keylen = (short int)dbkey.size;
        *inbuflen = dbvalue.size;
        memcpy(inbuf, dbvalue.data, dbvalue.size);
    }
    return stat;
}

int db_cursor_sk_find(int *it, struct dsc$descriptor *key, char *inbuf, int *maxinbuflen, int *inbuflen)
{
    int stat, key2len;
    char *key2;
    key2 = in_cstr(key);
    key2len = strlen(key2);
    stat = db_cursor_find(it, key2, &key2len, inbuf, maxinbuflen, inbuflen);
    free(key2);
    return stat;
}

int db_cursor_sk_next(int *it, struct dsc$descriptor *key, short int *keylen, char *inbuf, int *maxinbuflen, int *inbuflen)
{
    int maxkeylen;
    maxkeylen = key->dsc$w_length;
    return db_cursor_next(it, key->dsc$a_pointer, &maxkeylen, keylen, inbuf, maxinbuflen, inbuflen);
}

int db_cursor_ik_find(int *it, int *key, char *inbuf, int *maxinbuflen, int *inbuflen)
{
   return db_cursor_find(it, (char *)key, &size_of_int, inbuf, maxinbuflen, inbuflen);
}

int db_cursor_ik_next(int *it, int *key, char *inbuf, int *maxinbuflen, int *inbuflen)
{
   short int keylen;
   return db_cursor_next(it, (char *)key, &size_of_int, &keylen, inbuf, maxinbuflen, inbuflen);
}

void db_close(int *db)
{
    DB *db2;
    db2 = (DB *)(*db);
    db2->close(db2, 0);
}

